import logging
from typing import Any

from twisted.internet.defer import inlineCallbacks
from txhttputil.site.FileUnderlayResource import FileUnderlayResource
from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)
from vortex.handler.TupleDataObserverClient import TupleDataObserverClient

from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.PluginNames import searchObservableName
from peek_core_search._private.client.DeviceTupleProcessorActionProxy import (
    makeTupleActionProcessorProxy,
)
from peek_core_search._private.client.TupleDataObservable import (
    makeClientTupleDataObservableHandler,
)
from peek_core_search._private.client.controller.FastKeywordController import (
    FastKeywordController,
)
from peek_core_search._private.client.controller.SearchIndexCacheController import (
    SearchIndexCacheController,
)
from peek_core_search._private.client.controller.SearchObjectCacheController import (
    SearchObjectCacheController,
)
from peek_core_search._private.client.handlers.SearchIndexCacheHandler import (
    SearchIndexCacheHandler,
)
from peek_core_search._private.client.handlers.SearchObjectCacheHandler import (
    SearchObjectCacheHandler,
)
from peek_core_search._private.storage.DeclarativeBase import loadStorageTuples
from peek_core_search._private.tuples import loadPrivateTuples
from peek_core_search.tuples import loadPublicTuples
from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_base.client.PluginClientEntryHookABC import (
    PluginClientEntryHookABC,
)

logger = logging.getLogger(__name__)


class ClientEntryHook(PluginClientEntryHookABC):
    def __init__(self, *args, **kwargs):
        """ " Constructor"""
        # Call the base classes constructor
        PluginClientEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

    @classmethod
    def setupStaticWebResources(cls, platformApi: Any):
        from peek_plugin_base.client.PeekClientPlatformHookABC import (
            PeekClientPlatformHookABC,
        )

        assert isinstance(
            platformApi, PeekClientPlatformHookABC
        ), "We received the wrong platform API"

        from peek_abstract_chunked_index.private.client.controller.ACICacheControllerABC import (
            ACICacheControllerABC,
        )

        relativeDir = ACICacheControllerABC.appDownloadPluginDirRelativeDir()
        dirPath = platformApi.fileStorageDirectory / relativeDir
        dirPath.mkdir(exist_ok=True, parents=True)

        resource = FileUnderlayResource()
        resource.addFileSystemRoot(dirPath)

        platformApi.addFieldResource(relativeDir.encode(), resource)
        platformApi.addOfficeResource(relativeDir.encode(), resource)

    def load(self) -> None:
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """

        loadStorageTuples()

        loadPrivateTuples()
        loadPublicTuples()

        logger.debug("Loaded")

    @inlineCallbacks
    def start(self):
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        # ----------------

        # Provide the devices access to the servers observable
        tupleDataObservableProxyHandler = TupleDataObservableProxyHandler(
            observableName=searchObservableName,
            proxyToVortexName=peekServerName,
            additionalFilt=searchFilt,
            observerName="Proxy to devices",
        )
        self._loadedObjects.append(tupleDataObservableProxyHandler)

        # ----------------
        #: This is an observer for us (the client) to use to observe data
        # from the server
        serverTupleObserver = TupleDataObserverClient(
            observableName=searchObservableName,
            destVortexName=peekServerName,
            additionalFilt=searchFilt,
            observerName="Data for client",
        )
        self._loadedObjects.append(serverTupleObserver)

        # ----------------
        # Search Index Cache Controller

        searchIndexCacheController = SearchIndexCacheController(
            self.platform.serviceId, self.platform.fileStorageDirectory
        )
        self._loadedObjects.append(searchIndexCacheController)

        # This is the custom handler for the client
        searchIndexCacheHandler = SearchIndexCacheHandler(
            cacheController=searchIndexCacheController,
            clientId=self.platform.serviceId,
        )
        self._loadedObjects.append(searchIndexCacheHandler)

        searchIndexCacheController.setCacheHandler(searchIndexCacheHandler)

        # ----------------
        # Search Object Cache Controller

        searchObjectCacheController = SearchObjectCacheController(
            self.platform.serviceId, self.platform.fileStorageDirectory
        )
        self._loadedObjects.append(searchObjectCacheController)

        # This is the custom handler for the client
        searchObjectCacheHandler = SearchObjectCacheHandler(
            cacheController=searchObjectCacheController,
            clientId=self.platform.serviceId,
        )
        self._loadedObjects.append(searchObjectCacheHandler)

        searchObjectCacheController.setCacheHandler(searchObjectCacheHandler)

        # ----------------
        # Fast Keyword Controller

        fastKeywordController = FastKeywordController(
            searchObjectCacheController,
            searchIndexCacheController,
            serverTupleObserver,
        )
        self._loadedObjects.append(fastKeywordController)

        searchIndexCacheController.setFastKeywordController(
            fastKeywordController
        )

        # ----------------
        # Proxy actions back to the server, we don't process them at all
        tupleActionHandler = makeTupleActionProcessorProxy(
            fastKeywordController
        )
        self._loadedObjects.append(tupleActionHandler)

        # ----------------
        # Create the Tuple Observer
        makeClientTupleDataObservableHandler(
            tupleDataObservableProxyHandler,
            searchIndexCacheController,
            searchObjectCacheController,
        )

        # ----------------
        yield searchIndexCacheHandler.start()
        yield searchIndexCacheController.start()

        yield searchObjectCacheHandler.start()
        yield searchObjectCacheController.start()

        logger.debug("Started")

    def stop(self):
        """Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")



import logging

from vortex.Tuple import Tuple
from vortex.rpc.RPC import vortexRPC

from peek_abstract_chunked_index.private.server.client_handlers.ACIChunkLoadRpcABC import (
    ACIChunkLoadRpcABC,
)
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.storage.EncodedSearchIndexChunk import (
    EncodedSearchIndexChunk,
)
from peek_core_search._private.storage.EncodedSearchObjectChunk import (
    EncodedSearchObjectChunk,
)
from peek_core_search._private.tuples.search_index.SearchIndexUpdateDateTuple import (
    SearchIndexUpdateDateTuple,
)
from peek_core_search._private.tuples.search_object.SearchObjectUpdateDateTuple import (
    SearchObjectUpdateDateTuple,
)
from peek_plugin_base.PeekVortexUtil import peekBackendNames
from peek_plugin_base.PeekVortexUtil import peekServerName

logger = logging.getLogger(__name__)


class ClientChunkLoadRpc(ACIChunkLoadRpcABC):
    def makeHandlers(self):
        """Make Handlers

        In this method we start all the RPC handlers
        start() returns an instance of it's self so we can simply yield the result
        of the start method.

        """

        yield self.loadSearchIndexChunks.start(funcSelf=self)
        yield self.loadSearchIndexDelta.start(funcSelf=self)
        yield self.loadSearchObjectChunks.start(funcSelf=self)
        yield self.loadSearchObjectDelta.start(funcSelf=self)
        logger.debug("RPCs started")

    # -------------
    @vortexRPC(
        peekServerName,
        acceptOnlyFromVortex=peekBackendNames,
        timeoutSeconds=120,
        additionalFilt=searchFilt,
        deferToThread=True,
    )
    def loadSearchIndexDelta(self, indexEncodedPayload: bytes) -> bytes:
        return self.ckiChunkIndexDeltaBlocking(
            indexEncodedPayload,
            EncodedSearchIndexChunk,
            SearchIndexUpdateDateTuple,
        )

    # -------------
    @vortexRPC(
        peekServerName,
        acceptOnlyFromVortex=peekBackendNames,
        timeoutSeconds=120,
        additionalFilt=searchFilt,
        deferToThread=True,
    )
    def loadSearchIndexChunks(self, chunkKeys: list[str]) -> list[Tuple]:
        return self.ckiInitialLoadChunksPayloadBlocking(
            chunkKeys, EncodedSearchIndexChunk
        )

    # -------------
    @vortexRPC(
        peekServerName,
        acceptOnlyFromVortex=peekBackendNames,
        timeoutSeconds=120,
        additionalFilt=searchFilt,
        deferToThread=True,
    )
    def loadSearchObjectDelta(self, indexEncodedPayload: bytes) -> bytes:
        return self.ckiChunkIndexDeltaBlocking(
            indexEncodedPayload,
            EncodedSearchObjectChunk,
            SearchObjectUpdateDateTuple,
        )

    # -------------
    @vortexRPC(
        peekServerName,
        acceptOnlyFromVortex=peekBackendNames,
        timeoutSeconds=120,
        additionalFilt=searchFilt,
        deferToThread=True,
    )
    def loadSearchObjectChunks(self, chunkKeys: list[str]) -> list[Tuple]:
        return self.ckiInitialLoadChunksPayloadBlocking(
            chunkKeys, EncodedSearchObjectChunk
        )



from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_search._private.PluginNames import searchTuplePrefix


@addTupleType
class ExcludeSearchStringsTuple(Tuple):
    __tupleType__ = searchTuplePrefix + "ExcludeSearchStringsTuple"

    excludedPartialSearchTerms: list[str] = TupleField([])
    excludedFullSearchTerms: list[str] = TupleField([])



import logging
from typing import Dict
from typing import List

from twisted.internet.defer import inlineCallbacks

from peek_abstract_chunked_index.private.client.handlers.ACICacheHandlerABC import (
    ACICacheHandlerABC,
)
from peek_abstract_chunked_index.private.tuples.ACIUpdateDateTupleABC import (
    ACIUpdateDateTupleABC,
)
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.client.controller.SearchIndexCacheController import (
    SearchIndexCacheController,
)
from peek_core_search._private.client.controller.SearchIndexCacheController import (
    clientSearchIndexUpdateFromServerFilt,
)
from peek_core_search._private.tuples.search_index.SearchIndexUpdateDateTuple import (
    SearchIndexUpdateDateTuple,
)

logger = logging.getLogger(__name__)

clientSearchIndexWatchUpdateFromDeviceFilt = {
    "key": "clientSearchIndexWatchUpdateFromDevice"
}
clientSearchIndexWatchUpdateFromDeviceFilt.update(searchFilt)


# ModelSet HANDLER
class SearchIndexCacheHandler(ACICacheHandlerABC):
    _UpdateDateTuple: ACIUpdateDateTupleABC = SearchIndexUpdateDateTuple
    _updateFromDeviceFilt: Dict = clientSearchIndexWatchUpdateFromDeviceFilt
    _updateFromLogicFilt: Dict = clientSearchIndexUpdateFromServerFilt
    _logger: logging.Logger = logger

    @inlineCallbacks
    def notifyOfUpdate(self, chunkKeys: List[str]):
        assert isinstance(
            self._cacheController, SearchIndexCacheController
        ), "We expected SearchIndexCacheController"
        yield self._cacheController.notifyFastIndexOfChunkKeysUpdated(chunkKeys)
        yield ACICacheHandlerABC.notifyOfUpdate(self, chunkKeys)



from peek_core_search._private.PluginNames import searchTuplePrefix
from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class SearchResultOpenHandlerTuple(Tuple):
    """Search Result Tuple

    This tuple represents a search result
    """

    __tupleType__ = searchTuplePrefix + "SearchResultOpenHandlerTuple"

    #:  The key of a registered open handler
    key: str = TupleField()

    #:  The name of the open handler
    name: str = TupleField()

    #:  The description of the open handlers action type
    title: str = TupleField()



import logging

from peek_core_search._private.PluginNames import searchTuplePrefix
from sqlalchemy import Column, BigInteger
from sqlalchemy import Integer, String
from sqlalchemy.sql.schema import Index, ForeignKey

from vortex.Tuple import Tuple, addTupleType
from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class SearchIndex(Tuple, DeclarativeBase):
    __tablename__ = "SearchIndex"
    __tupleType__ = searchTuplePrefix + "SearchIndexTable"

    id = Column(BigInteger, primary_key=True, autoincrement=True)

    chunkKey = Column(String, nullable=False)
    keyword = Column(String, nullable=False)
    propertyName = Column(String, nullable=False)

    #:  The object that this routs is for
    objectId = Column(
        BigInteger,
        ForeignKey("SearchObject.id", ondelete="CASCADE"),
        nullable=False,
    )

    __table_args__ = (
        Index(
            "idx_SearchIndex_quick_query",
            chunkKey,
            keyword,
            propertyName,
            objectId,
            unique=True,
        ),
        Index("idx_SearchIndex_objectId", objectId),  # For foreign key
    )



"""added show to prop

Peek Plugin Database Migration Script

Revision ID: 36aeb6eaea5c
Revises: c18d0a2ab678
Create Date: 2019-07-30 21:10:50.837891

"""

# revision identifiers, used by Alembic.
revision = "36aeb6eaea5c"
down_revision = "c18d0a2ab678"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ###
    op.add_column(
        "SearchProperty",
        sa.Column("showInHeader", sa.Boolean(), server_default="0", nullable=True),
        schema="core_search",
    )

    op.execute(""" UPDATE "core_search"."SearchProperty" SET "showInHeader" = false """)

    op.alter_column(
        "SearchProperty",
        "showInHeader",
        type_=sa.Boolean(),
        nullable=False,
        schema="core_search",
    )

    # ###
    op.add_column(
        "SearchProperty",
        sa.Column("showOnResult", sa.Boolean(), server_default="1", nullable=True),
        schema="core_search",
    )

    op.execute(""" UPDATE "core_search"."SearchProperty" SET "showOnResult" = true """)

    op.alter_column(
        "SearchProperty",
        "showOnResult",
        type_=sa.Boolean(),
        nullable=False,
        schema="core_search",
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("SearchProperty", "showOnResult", schema="core_search")
    op.drop_column("SearchProperty", "showInHeader", schema="core_search")
    # ### end Alembic commands ###
